/*
**
**
**  VIA Technologies, Inc.              
**
**  Fast Ethernet Adapter
**
**  Linux Driver
**
**  v3.23  May. 2001
**
**
*/

/* These identify the driver base version and may not be removed. */
static const char version[] =
"linuxfet.c : v3.23 05/15/2001\n";

#include "linuxfet.h"
static int debug = 3;			/* 1 normal messages, 0 quiet .. 7 verbose. */
static int min_pci_latency = 32;

/* The user-configurable values.
   These may be modified when a driver module is loaded.*/   
static int max_interrupt_work = 20;

/* Set the copy breakpoint for the copy-only-tiny-frames scheme.
   Setting to > 1518 effectively disables this feature. */
static int rx_copybreak = 0;

/* Used to pass the media type, etc.
   Both 'options[]' and 'full_duplex[]' should exist for driver
   interoperability.
   The media type is usually passed in 'options[]'.
*/
#define MAX_UNITS 8     /* More are supported, limit only on options */
static int options[MAX_UNITS] = {-1, -1, -1, -1, -1, -1, -1, -1};
static int full_duplex[MAX_UNITS] = {-1, -1, -1, -1, -1, -1, -1, -1};
/* Maximum number of multicast addresses to filter (vs. rx-all-multicast).
   The Rhine has a 64 element 8390-like hash table.  */
static const int multicast_filter_limit = 32;

pci_id_info pci_tbl[] = {
	{ "VIA VT86C100A Fast Ethernet Adapter                         ", 
	  { 0x30431106, 0xffffffff, },
	  RHINE_IOTYPE, 128, CanHaveMII | ReqTxAlign | HasDavicomPhy},
	{ "VIA PCI 10/100Mb Fast Ethernet Adapter                      ",
          { 0x30651106, 0xffffffff, },
	  RHINE_IOTYPE, RHINEII_IOSIZE, CanHaveMII | HasWOL },
	{0,},						/* 0 terminated list. */
};


drv_id_info via_rhine_drv_id = {
    "linuxfet", 0, PCI_CLASS_NETWORK_ETHERNET<<8, pci_tbl, via_probe1,};
/* Offsets to the device registers.
*/


#if LINUX_VERSION_CODE > 0x20118  &&  defined(MODULE)
MODULE_AUTHOR("Donald Becker <becker@scyld.com>");
MODULE_DESCRIPTION("PCI 10/100Mb Fast Ethernet Adapter");
MODULE_PARM(debug, "i");
MODULE_PARM(min_pci_latency, "i");
MODULE_PARM(max_interrupt_work, "i");
MODULE_PARM(rx_copybreak, "i");
MODULE_PARM(options, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(full_duplex, "1-" __MODULE_STRING(MAX_UNITS) "i");
#endif



/* A list of our installed devices, for removing the driver module. */
static struct net_device *root_net_dev = NULL;
#if LINUX_VERSION_CODE >= 0x20200
/* Grrrr.. complex abstaction layers with negative benefit. */
static int pci_drv_register(drv_id_info *drv_id, void *initial_device)
{
	int chip_idx, cards_found = 0;
	struct pci_dev *pdev = NULL;
	pci_id_info *pci_tbl = drv_id->pci_dev_tbl;
	void *newdev;
        u32 revision;

	while ((pdev = pci_find_class(drv_id->pci_class, pdev)) != 0) {
		u32 pci_id, pci_subsys_id, pci_class_rev;
		u16 pci_command, new_command;
		int pci_flags;
		long pciaddr;			/* Bus address. */
		long ioaddr;			/* Mapped address for this processor. */
                
		pci_read_config_dword(pdev, PCI_VENDOR_ID, &pci_id);
		/* Offset 0x2c is PCI_SUBSYSTEM_ID aka PCI_SUBSYSTEM_VENDOR_ID. */
		pci_read_config_dword(pdev, 0x2c, &pci_subsys_id);
		pci_read_config_dword(pdev, PCI_REVISION_ID, &pci_class_rev);
                
                revision = (pci_class_rev << 16) >> 16;
		if (debug > 3)
			printk(KERN_DEBUG "PCI ID %8.8x subsystem ID is %8.8x.\n",
				   pci_id, pci_subsys_id);
		for (chip_idx = 0; pci_tbl[chip_idx].name; chip_idx++) {
			pci_id_info *chip = &pci_tbl[chip_idx];
			if ((pci_id & chip->id.pci_mask) == chip->id.pci
				&& (pci_subsys_id&chip->id.subsystem_mask) == chip->id.subsystem
				&& (pci_class_rev&chip->id.revision_mask) == chip->id.revision)
				break;
		}
		if (pci_tbl[chip_idx].name == 0) 		/* Compiled out! */
			continue;

		pci_flags = pci_tbl[chip_idx].pci_flags;
#if (LINUX_VERSION_CODE >= 0x02030c)
		/* Wow. A oversized, hard-to-use abstraction. Bogus. */
		pciaddr = pdev->resource[(pci_flags >> 4) & 7].start;
                ioaddr = pci_resource_start(pdev, 0);

#else
		pciaddr = pdev->base_address[(pci_flags >> 4) & 7];
#if defined(__alpha__)			/* Really any machine with 64 bit addressing. */
		if (pci_flags & PCI_ADDR_64BITS)
			pciaddr |= ((long)pdev->base_address[((pci_flags>>4)&7)+ 1]) << 32;
#endif

		if (debug > 2)
			printk(KERN_INFO "Found %s at PCI address %#lx, mapped IRQ %d.\n",
				   pci_tbl[chip_idx].name, pciaddr, pdev->irq);

		if ((pciaddr & PCI_BASE_ADDRESS_SPACE_IO)) {
			ioaddr = pciaddr & PCI_BASE_ADDRESS_IO_MASK;
			if (check_region(ioaddr, pci_tbl[chip_idx].io_size))
				continue;
		} else if ((ioaddr = (long)ioremap(pciaddr & PCI_BASE_ADDRESS_MEM_MASK,
										   pci_tbl[chip_idx].io_size)) == 0) {
			printk(KERN_INFO "Failed to map PCI address %#lx for device "
				   "'%s'.\n", pciaddr, pci_tbl[chip_idx].name);
			continue;
		}
#endif
		if ( ! (pci_flags & PCI_NO_ACPI_WAKE))
			acpi_wake(pdev);
		pci_read_config_word(pdev, PCI_COMMAND, &pci_command);
		new_command = pci_command | (pci_flags & 7);
		if (pci_command != new_command) {
			printk(KERN_INFO "  The PCI BIOS has not enabled the"
				   " device at %d/%d!  Updating PCI command %4.4x->%4.4x.\n",
				   pdev->bus->number, pdev->devfn, pci_command, new_command);
			pci_write_config_word(pdev, PCI_COMMAND, new_command);
		}

		newdev = drv_id->probe1(pdev, initial_device,
								ioaddr, pdev->irq, chip_idx, cards_found, revision);
		if (newdev  && (pci_flags & PCI_COMMAND_MASTER))
			pci_set_master(pdev);
		if (newdev  && (pci_flags & PCI_COMMAND_MASTER)  &&
			! (pci_flags & PCI_NO_MIN_LATENCY)) {
			u8 pci_latency;
			pci_read_config_byte(pdev, PCI_LATENCY_TIMER, &pci_latency);
			if (pci_latency < min_pci_latency) {
				printk(KERN_INFO "  PCI latency timer (CFLT) is "
					   "unreasonably low at %d.  Setting to %d clocks.\n",
					   pci_latency, min_pci_latency);
				pci_write_config_byte(pdev, PCI_LATENCY_TIMER, min_pci_latency);
			}
		}
		initial_device = 0;
		cards_found++;
	}

	if ((drv_id->flags & PCI_HOTSWAP)
		&& register_cb_hook
		&& (*register_cb_hook)(drv_id) == 0) {
		MOD_INC_USE_COUNT;
		return 0;
	} else
		return cards_found ? 0 : -ENODEV;
}
#else
static int pci_drv_register(drv_id_info *drv_id, void *initial_device)
{
	int pci_index, cards_found = 0;
	unsigned char pci_bus, pci_device_fn;
	struct pci_dev *pdev;
	pci_id_info *pci_tbl = drv_id->pci_dev_tbl;
	void *newdev;
        u32 revision;

	if ( ! pcibios_present())
		return -ENODEV;

	for (pci_index = 0; pci_index < 0xff; pci_index++) {
		u32 pci_id, subsys_id, pci_class_rev;
		u16 pci_command, new_command;
		int chip_idx, irq, pci_flags;
		long pciaddr;
		long ioaddr;
		u32 pci_busaddr;
		u8 pci_irq_line;

		if (pcibios_find_class (drv_id->pci_class, pci_index,
								&pci_bus, &pci_device_fn)
			!= PCIBIOS_SUCCESSFUL)
			break;
		pcibios_read_config_dword(pci_bus, pci_device_fn,
								  PCI_VENDOR_ID, &pci_id);
		/* Offset 0x2c is PCI_SUBSYSTEM_ID aka PCI_SUBSYSTEM_VENDOR_ID. */
		pcibios_read_config_dword(pci_bus, pci_device_fn, 0x2c, &subsys_id);
		pcibios_read_config_dword(pci_bus, pci_device_fn,
								  PCI_REVISION_ID, &pci_class_rev);

		for (chip_idx = 0; pci_tbl[chip_idx].name; chip_idx++) {
			pci_id_info *chip = &pci_tbl[chip_idx];
			if ((pci_id & chip->id.pci_mask) == chip->id.pci
				&& (subsys_id & chip->id.subsystem_mask) == chip->id.subsystem
				&& (pci_class_rev&chip->id.revision_mask) == chip->id.revision)
				break;
		}
		if (pci_tbl[chip_idx].name == 0) 		/* Compiled out! */
			continue;

		pci_flags = pci_tbl[chip_idx].pci_flags;
		pdev = pci_find_slot(pci_bus, pci_device_fn);
		pcibios_read_config_byte(pci_bus, pci_device_fn,
								 PCI_INTERRUPT_LINE, &pci_irq_line);
		irq = pci_irq_line;
		pcibios_read_config_dword(pci_bus, pci_device_fn,
								  ((pci_flags >> 2) & 0x1C) + 0x10,
								  &pci_busaddr);
		pciaddr = pci_busaddr;
#if defined(__alpha__)
		if (pci_flags & PCI_ADDR_64BITS) {
			pcibios_read_config_dword(pci_bus, pci_device_fn,
									  ((pci_flags >> 2) & 0x1C) + 0x14,
									  &pci_busaddr);
			pciaddr |= ((long)pci_busaddr)<<32;
		}
#endif

		if (debug > 2)
			printk(KERN_INFO "Found %s at PCI address %#lx, IRQ %d.\n",
				   pci_tbl[chip_idx].name, pciaddr, irq);

		if ((pciaddr & PCI_BASE_ADDRESS_SPACE_IO)) {
			ioaddr = pciaddr & PCI_BASE_ADDRESS_IO_MASK;
			if (check_region(ioaddr, pci_tbl[chip_idx].io_size))
				continue;
		} else if ((ioaddr = (long)ioremap(pciaddr & PCI_BASE_ADDRESS_MEM_MASK,
										   pci_tbl[chip_idx].io_size)) == 0) {
			printk(KERN_INFO "Failed to map PCI address %#lx.\n",
				   pciaddr);
			continue;
		}

		if ( ! (pci_flags & PCI_NO_ACPI_WAKE))
			acpi_wake(pdev);
		pcibios_read_config_word(pci_bus, pci_device_fn,
								 PCI_COMMAND, &pci_command);
		new_command = pci_command | (pci_flags & 7);
		if (pci_command != new_command) {
			printk(KERN_INFO "  The PCI BIOS has not enabled the"
				   " device at %d/%d!  Updating PCI command %4.4x->%4.4x.\n",
				   pci_bus, pci_device_fn, pci_command, new_command);
			pcibios_write_config_word(pci_bus, pci_device_fn,
									  PCI_COMMAND, new_command);
		}

		newdev = drv_id->probe1(pdev, initial_device,
							   ioaddr, irq, chip_idx, cards_found, revision);

		if (newdev  && (pci_flags & PCI_COMMAND_MASTER)  &&
			! (pci_flags & PCI_NO_MIN_LATENCY)) {
			u8 pci_latency;
			pcibios_read_config_byte(pci_bus, pci_device_fn,
									 PCI_LATENCY_TIMER, &pci_latency);
			if (pci_latency < min_pci_latency) {
				printk(KERN_INFO "  PCI latency timer (CFLT) is "
					   "unreasonably low at %d.  Setting to %d clocks.\n",
					   pci_latency, min_pci_latency);
				pcibios_write_config_byte(pci_bus, pci_device_fn,
										  PCI_LATENCY_TIMER, min_pci_latency);
			}
		}
		initial_device = 0;
		cards_found++;
	}

	if ((drv_id->flags & PCI_HOTSWAP)
		&& register_cb_hook
		&& (*register_cb_hook)(drv_id) == 0) {
		MOD_INC_USE_COUNT;
		return 0;
	} else
		return cards_found ? 0 : -ENODEV;
}
#endif

static void pci_drv_unregister(drv_id_info *drv_id)
{
	/* We need do something only with CardBus support. */
	if (unregister_cb_hook) {
		(*unregister_cb_hook)(drv_id);
		MOD_DEC_USE_COUNT;
	}
	return;
}

/*
  Search PCI configuration space for the specified capability registers.
  Return the index, or 0 on failure.
*/
static int pci_find(struct pci_dev *pdev, int findtype)
{
	u16 pci_status, cap_type;
	u8 pci_cap_idx;
	int cap_idx;

	pci_read_config_word(pdev, PCI_STATUS, &pci_status);
	if ( ! (pci_status & PCI_STATUS_CAP_LIST))
		return 0;
	pci_read_config_byte(pdev, PCI_CAPABILITY_LIST, &pci_cap_idx);
	cap_idx = pci_cap_idx;
	for (cap_idx = pci_cap_idx; cap_idx; cap_idx = (cap_type >> 8) & 0xff) {
		pci_read_config_word(pdev, cap_idx, &cap_type);
		if ((cap_type & 0xff) == findtype)
			return cap_idx;
	}
	return 0;
}


/* Change a device from D3 (sleep) to D0 (active).
   Return the old power state.
   This is more complicated than you might first expect since most cards
   forget all PCI config info during the transition! */
static int acpi_wake(struct pci_dev *pdev)
{
	u32 base[5], romaddr;
	u16 pci_command, pwr_command;
	u8  pci_latency, pci_cacheline, irq;
	int i, pwr_cmd_idx = pci_find(pdev, PCI_CAP_ID_PM);

	if (pwr_cmd_idx == 0)
		return 0;
	pci_read_config_word(pdev, pwr_cmd_idx + 4, &pwr_command);
	if ((pwr_command & 3) == 0)
		return 0;
	pci_read_config_word(pdev, PCI_COMMAND, &pci_command);
	for (i = 0; i < 5; i++)
		pci_read_config_dword(pdev, PCI_BASE_ADDRESS_0 + i*4,
								  &base[i]);
	pci_read_config_dword(pdev, PCI_ROM_ADDRESS, &romaddr);
	pci_read_config_byte( pdev, PCI_LATENCY_TIMER, &pci_latency);
	pci_read_config_byte( pdev, PCI_CACHE_LINE_SIZE, &pci_cacheline);
	pci_read_config_byte( pdev, PCI_INTERRUPT_LINE, &irq);

	pci_write_config_word(pdev, pwr_cmd_idx + 4, 0x0000);
	for (i = 0; i < 5; i++)
		if (base[i])
			pci_write_config_dword(pdev, PCI_BASE_ADDRESS_0 + i*4,
									   base[i]);
	pci_write_config_dword(pdev, PCI_ROM_ADDRESS, romaddr);
	pci_write_config_byte( pdev, PCI_INTERRUPT_LINE, irq);
	pci_write_config_byte( pdev, PCI_CACHE_LINE_SIZE, pci_cacheline);
	pci_write_config_byte( pdev, PCI_LATENCY_TIMER, pci_latency);
	pci_write_config_word( pdev, PCI_COMMAND, pci_command | 5);
	return pwr_command & 3;
}

static void *via_probe1(struct pci_dev *pdev, void *init_dev,
                        long ioaddr, int irq, int chip_idx, int card_idx, u32 revision)
{
    struct net_device *dev;
    netdev_private *np;
    void *priv_mem;
    int i, option = card_idx < MAX_UNITS ? options[card_idx] : 0;
    unsigned char byOrgValue;
    int ww;
    u8 mode3_reg;

    // if vt3065     
    if (revision>=0x40) {
       // clear sticky bit before reset & read ethernet address
       byOrgValue = readb(ioaddr + StickyHW);    
       byOrgValue = byOrgValue & 0xFC;
       writeb(byOrgValue, ioaddr + StickyHW);        
       // disable force PME-enable 
       writeb(0x80, ioaddr + WOLcgClr);
       // disable power-event config bit
       writeb(0xFF, ioaddr + WOLcrClr);
       // clear power status 
       writeb(0xFF, ioaddr + PwrcsrClr);   
    }
    
        
    /* Reset the chip to erase previous misconfiguration. */
    writew(CmdReset, ioaddr + ChipCmd);
    // if vt3043 delay after reset
    if (revision <0x40) {
       udelay(100);
    }

    // polling till software reset complete
    // W_MAX_TIMEOUT is the timeout period
    for (ww = 0; ww < W_MAX_TIMEOUT; ww++) {
        if ((readw(ioaddr + ChipCmd) & CmdReset) == 0 )
            break;
    }
       
    // issue AUTOLoad in EECSR to reload eeprom
    writeb(0x20, ioaddr + MACRegEEcsr);
      
    // if vt3065 delay after reset
    if (revision >=0x40 ) { 

        // delay 8ms to let MAC stable
        mdelay(8);
        // for 3065D, EEPROM reloaded will cause bit 0 in MAC_REG_CFGA turned on.
        // it makes MAC receive magic packet automatically. So, driver turn it off.
        writeb(readb(ioaddr + ConfigA) & 0xFE, ioaddr + ConfigA);
    }

    /* turn on bit2 in PCI configuration register 0x53 , only for 3065*/
    if (revision >= 0x40) {
        pci_read_config_byte(pdev,PCI_REG_MODE3,&mode3_reg);
        pci_write_config_byte(pdev,PCI_REG_MODE3,mode3_reg|MODE3_MIION);
    }

    /* back off algorithm ,disable the right-most 4-bit off CFGD*/
    writeb(readb(ioaddr + ConfigD) & (~(CRADOM | CAP | MBA | BAKOPT)), ioaddr + ConfigD);

    dev = init_etherdev(init_dev, 0);

    printk(KERN_INFO "%s: %s\n",dev->name, pci_tbl[chip_idx].name);
    printk(KERN_INFO "%s: IO Address = 0x%lx, MAC Address = ",dev->name, ioaddr);
    /* Ideally we would read the EEPROM but access may be locked. */
    for (i = 0; i < 6; i++)
        dev->dev_addr[i] = readb(ioaddr + StationAddr + i);
    for (i = 0; i < 5; i++)
            printk("%2.2x:", dev->dev_addr[i]);
    printk("%2.2x, IRQ = %d.\n", dev->dev_addr[i], irq);

    /* Make certain the descriptor lists are cache-aligned. */
    priv_mem = kmalloc(sizeof(*np) + PRIV_ALIGN, GFP_KERNEL);
    /* Check for the very unlikely case of no memory. */
    if (priv_mem == NULL)
        return NULL;

#ifdef USE_IO_OPS
    request_region(ioaddr, pci_tbl[chip_idx].io_size, dev->name);
#endif

    /* Reset the chip to erase previous misconfiguration. */
//    writew(CmdReset, ioaddr + ChipCmd);

    dev->base_addr = ioaddr;
    dev->irq = irq;

    dev->priv = np = (void *)(((long)priv_mem + PRIV_ALIGN) & ~PRIV_ALIGN);
    memset(np, 0, sizeof(*np));
    np->priv_addr = priv_mem;
    np->next_module = root_net_dev;
    root_net_dev = dev;

    np->pci_dev = pdev;
    np->chip_id = chip_idx;
    np->drv_flags = pci_tbl[chip_idx].drv_flags;

    if (dev->mem_start)
        option = dev->mem_start;
        
       
    /* The lower four bits are the media type. */
    if (option > 0) {
        if (option & 0x200)
            np->full_duplex = 1;
        np->default_port = option & 15;
        if (np->default_port)
            np->medialock = 1;
    }
    if (card_idx < MAX_UNITS  &&  full_duplex[card_idx] > 0)
        np->full_duplex = 1;

    if (np->full_duplex)
        np->duplex_lock = 1;

    /* The chip-specific entries in the device structure. */
    dev->open = &netdev_open;
    dev->hard_start_xmit = &start_tx;
    dev->stop = &netdev_close;
    dev->get_stats = &get_stats;
    dev->set_multicast_list = &set_rx_mode;
    dev->do_ioctl = &mii_ioctl;

    if (np->drv_flags & CanHaveMII) {
        int phy, phy_idx = 0;
        np->phys[0] = 1;        /* Standard for this chip. */
        for (phy = 1; phy < 32 && phy_idx < 4; phy++) {
            int mii_status = mdio_read(dev, phy, 1);
            if (mii_status != 0xffff  &&  mii_status != 0x0000) {
                np->phys[phy_idx++] = phy;
                np->advertising = mdio_read(dev, phy, 4);
                printk(KERN_INFO "%s: MII PHY found at address %d, status "
                       "0x%4.4x advertising %4.4x Link %4.4x.\n",
                       dev->name, phy, mii_status, np->advertising,
                       mdio_read(dev, phy, 5));
            }
        }
        np->mii_cnt = phy_idx;
    }

    return dev;
}

/* Read and write over the MII Management Data I/O (MDIO) interface. */
static int mdio_read(struct net_device *dev, int phy_id, int regnum)
{
    long ioaddr = dev->base_addr;
    int boguscnt = 0x3fff;
    char byMIIAdrbak;
    char byMIICRbak;
    unsigned wMIIData;
    netdev_private *np = (netdev_private *)dev->priv;

    /* backup MIIAddr & MIICmd*/
    byMIIAdrbak = readb(ioaddr + MIIRegAddr);
    byMIICRbak = readb(ioaddr + MIICmd);
    
    /* before read mii data, we must turn off mauto */
    writeb(byMIICRbak & 0x7f, ioaddr + MIICmd);
    
    // After turn off mauto, we must make sure mauto is off 
    if (np->drv_flags & HasDavicomPhy)  {  
        //if 3043, loop delay
        udelay(100);
    }
    else {                   
        //if 3065, waiting midle
        while (!(readb(ioaddr + MIIRegAddr) & 0x80) && --boguscnt > 0)
        ;
    }
    /* write phy id */
    writeb(phy_id, ioaddr + MIIPhyAddr);
    /* write mii address */
    writeb(regnum, ioaddr + MIIRegAddr);
    /* Trigger read */
    writeb(readb(ioaddr+MIICmd) | 0x40, ioaddr + MIICmd);   
    boguscnt = 0x3fff;
    
    /* waiting read complete. */
    while ((readb(ioaddr + MIICmd) & 0x40) && --boguscnt > 0)
        ;
    udelay(100);    
    /* read mii data */    
    wMIIData = readw(ioaddr + MIIData);
        
    /* restore MII Data & turn on mauto*/
    writeb(byMIIAdrbak, ioaddr + MIIRegAddr);
    writeb(byMIICRbak, ioaddr + MIICmd);
    
    return wMIIData;        
}

static void mdio_write(struct net_device *dev, int phy_id, int regnum, int value)
{
    long ioaddr = dev->base_addr;
    int boguscnt = 0x3fff;
    char byMIIAdrbak;
    char byMIICRbak;
    netdev_private *np = (netdev_private *)dev->priv;
    
    
    if (phy_id == np->phys[0]) {
		switch (regnum) {
		case 0:					/* Is user forcing speed/duplex? */
			if (value & 0x9000)	/* Autonegotiation. */
				np->duplex_lock = 0;
			else
				np->full_duplex = (value & 0x0100) ? 1 : 0;
			break;
		case 4: np->advertising = value; break;
		}
	}

    /* backup MIIAddr & MIICmd*/
    byMIIAdrbak = readb(ioaddr + MIIRegAddr);
    byMIICRbak = readb(ioaddr + MIICmd);

    /* before read mii data, we must turn off mauto */
    writeb(byMIICRbak & 0x7f, ioaddr + MIICmd);
    
    /* before write mii data, we must turn off mauto */
    if (np->drv_flags & HasDavicomPhy)  {  
        //if 3043, loop delay
        udelay(100);
    }
    else {                   
        //if 3071, waiting midle
        while (!(readb(ioaddr + MIIRegAddr) & 0x80) && --boguscnt > 0)
                  ;
    }
    /* write phy id */
    writeb(phy_id, ioaddr + MIIPhyAddr);
    /* write mii address */
    writeb(regnum, ioaddr + MIIRegAddr);
    /* write mii data */
    writew(value, ioaddr + MIIData);
    /* Trigger write */
    writeb(readb(ioaddr+MIICmd) | 0x20, ioaddr+MIICmd);
    /* waiting write complete. */
    while ((readb(ioaddr + MIICmd) & 0x20) && --boguscnt > 0)
        ;
        
    /* restore MII Data & turn on mauto*/
    writeb(byMIIAdrbak, ioaddr + MIIRegAddr);
    writeb(byMIICRbak, ioaddr + MIICmd);
    
    return;
}


static int netdev_open(struct net_device *dev)
{
    netdev_private *np = (netdev_private *)dev->priv;
    long ioaddr = dev->base_addr;
    int i, ww ;
    int phy_id1, phy_id2;

    /* Reset the chip. */
    writew(CmdReset, ioaddr + ChipCmd);
    // polling till software reset complete
    // W_MAX_TIMEOUT is the timeout period
    for (ww = 0; ww < W_MAX_TIMEOUT; ww++) {
        if ((readw(ioaddr + ChipCmd) & CmdReset) == 0 )
            break;
    }

    MOD_INC_USE_COUNT;

    if (request_irq(dev->irq, &intr_handler, SA_SHIRQ, dev->name, dev)) {

        MOD_DEC_USE_COUNT;
        return -EAGAIN;
    }

    if (debug > 1)
        printk(KERN_INFO "%s: netdev_open() irq %d.\n",
               dev->name, dev->irq);

    init_ring(dev);

    writel(virt_to_bus(np->rx_ring), ioaddr + RxRingPtr);
    writel(virt_to_bus(np->tx_ring), ioaddr + TxRingPtr);

    for (i = 0; i < 6; i++)
        writeb(dev->dev_addr[i], ioaddr + StationAddr + i);

    /* Initialize other registers. */    
    // Turn on bit3 (OFSET) in TCR during MAC initialization.
    writeb(readb(ioaddr + TxConfig) | TCR_OFSET, ioaddr + TxConfig);
    
    // turn on mauto
    writeb(0x80, ioaddr + MIICmd);
    writeb(0x41, ioaddr + MIIRegAddr);
    writew(0x0006, ioaddr + PCIBusConfig);    /* Tune configuration??? */
    /* Configure the FIFO thresholds. */
    writeb(readb(ioaddr + TxConfig) | 0x80, ioaddr + TxConfig);    /* Initial Tx threshold 32 bytes */ 
    np->tx_thresh = 0x20;
    np->rx_thresh = 0xE0;               /* Initial Rx threshold store & forward */
                                            /* Written in set_rx_mode(). */

    if (dev->if_port == 0)
        dev->if_port = np->default_port;

#if (LINUX_VERSION_CODE >= 0x02032a)
    netif_start_queue(dev);
#else
    dev->tbusy = 0;
    dev->interrupt = 0;
    dev->start = 1;
#endif

    set_rx_mode(dev);

    /* Enable interrupts by setting the interrupt mask. */
    writew(IntrRxDone | IntrRxErr | IntrRxEmpty| IntrRxOverflow| IntrRxDropped|
           IntrTxDone | IntrTxAbort | IntrTxUnderrun |
           IntrPCIErr | IntrStatsMax | IntrLinkChange | IntrMIIChange,
           ioaddr + IntrEnable);

    np->chip_cmd = CmdStart|CmdTxOn|CmdRxOn|CmdNoTxPoll;
    if (np->duplex_lock)
        np->chip_cmd |= CmdFDuplex;
    writew(np->chip_cmd, ioaddr + ChipCmd);
    

    /* restart MII auto-negotiation for davidcom phy bug*/
    phy_id1 = mdio_read(dev, np->phys[0], 0x02);
    phy_id2 = mdio_read(dev, np->phys[0], 0x03);  
    
    /* if phy vendor id == davidcom phy */        
    if (phy_id1 ==  0x0181 && phy_id2 == 0xb800) {
        mdio_write(dev,np->phys[0], 0 ,mdio_read(dev, np->phys[0], 0) | 0x1200);
    }    

    check_duplex(dev);
    
    
    /* The LED outputs of various MII xcvrs should be configured.  */
    /* For NS or Mison phys, turn on bit 1 in register 0x17 */
    /* For ESI phys, turn on bit 7 in register 0x17. */
    mdio_write(dev, np->phys[0], 0x17, mdio_read(dev, np->phys[0], 0x17) |
               (np->drv_flags & HasESIPhy) ? 0x0080 : 0x0001);

    if (debug > 2)
        printk(KERN_INFO "%s: Done netdev_open(), status %4.4x "
               "MII status: %4.4x.\n",
               dev->name, readw(ioaddr + ChipCmd),
               mdio_read(dev, np->phys[0], 1));

    /* Set the timer to check for link beat. */
    init_timer(&np->timer);
    np->timer.expires = jiffies + 2;
    np->timer.data = (unsigned long)dev;
    np->timer.function = &netdev_timer;             /* timer handler */
    add_timer(&np->timer);
    return 0;

}

static void check_duplex(struct net_device *dev)
{
    netdev_private *np = (netdev_private *)dev->priv;
    long ioaddr = dev->base_addr;
    int mii_reg5 = mdio_read(dev, np->phys[0], 5);
    int negotiated = mii_reg5 & np->advertising;
    int duplex;

    if (np->duplex_lock  ||  mii_reg5 == 0xffff)
        return;
    duplex = (negotiated & 0x0100) || (negotiated & 0x01C0) == 0x0040;
    if (np->full_duplex != duplex) {
        np->full_duplex = duplex;
        if (debug)
            printk(KERN_INFO "%s: Setting %s-duplex based on MII #%d link"
                   " partner capability of %4.4x.\n", dev->name,
                   duplex ? "full" : "half", np->phys[0], mii_reg5);
        if (duplex)
            np->chip_cmd |= CmdFDuplex;
        else
            np->chip_cmd &= ~CmdFDuplex;
        writew(np->chip_cmd, ioaddr + ChipCmd);
    }
}
        
static void netdev_timer(unsigned long data)
{
    struct net_device *dev = (struct net_device *)data;
    netdev_private *np = (netdev_private *)dev->priv;
    int next_tick = 10*HZ;

#if (LINUX_VERSION_CODE >= 0x02032a)
    if (netif_queue_stopped(dev) != 0
        && np->cur_tx - np->dirty_tx > 1
        && jiffies - dev->trans_start > TX_TIMEOUT)
#else
    if (test_bit(0, (void*)&dev->tbusy) != 0
        && np->cur_tx - np->dirty_tx > 1
        && jiffies - dev->trans_start > TX_TIMEOUT)
#endif
        tx_timeout(dev);
    // link change will generate interrupt so we must not check_duplex in timer
    //check_duplex(dev);
    

    np->timer.expires = jiffies + next_tick;
    add_timer(&np->timer);
}

static void tx_timeout(struct net_device *dev)
{
    netdev_private *np = (netdev_private *)dev->priv;
    long ioaddr = dev->base_addr;


    printk(KERN_INFO "%s: Transmit timed out, status %4.4x, PHY status "
           "%4.4x, resetting...\n",
           dev->name, readw(ioaddr + IntrStatus),
           mdio_read(dev, np->phys[0], 1));

    /* Perhaps we should reinitialize the hardware here. */
    dev->if_port = 0;
    /* Stop and restart the chip's Tx processes . */
    netdev_close(dev);
    netdev_open(dev);
        
    /* Trigger an immediate transmit demand. */
    dev->trans_start = jiffies;
    np->stats.tx_errors++;
    return;
}


/* Initialize the Rx and Tx rings, along with various 'dev' bits. */
static void init_ring(struct net_device *dev)
{
    netdev_private *np = (netdev_private *)dev->priv;
    int i;

    np->tx_full = 0;
    np->cur_rx = np->cur_tx = 0;
    np->dirty_rx = np->dirty_tx = 0;

    np->rx_buf_sz = (dev->mtu <= 1500 ? PKT_BUF_SZ : dev->mtu + 32);
    np->rx_head_desc = &np->rx_ring[0];

    for (i = 0; i < RX_RING_SIZE; i++) {
        np->rx_ring[i].rx_status = 0;
        np->rx_ring[i].desc_length = cpu_to_le32(np->rx_buf_sz);
        np->rx_ring[i].next_desc = virt_to_le32desc(&np->rx_ring[i+1]);
        np->rx_skbuff[i] = 0;
    }
    /* Mark the last entry as wrapping the ring. */
    np->rx_ring[i-1].next_desc = virt_to_le32desc(&np->rx_ring[0]);

    /* Fill in the Rx buffers.  Handle allocation failure gracefully. */
    for (i = 0; i < RX_RING_SIZE; i++) {
        struct sk_buff *skb = dev_alloc_skb(np->rx_buf_sz);
        np->rx_skbuff[i] = skb;
        if (skb == NULL)
            break;
        skb->dev = dev;         /* Mark as being used by this device. */
        np->rx_ring[i].addr = virt_to_le32desc(skb->tail);
        np->rx_ring[i].rx_status = cpu_to_le32(DescOwn);
    }
    np->dirty_rx = (unsigned int)(i - RX_RING_SIZE);

    for (i = 0; i < TX_RING_SIZE; i++) {
        np->tx_skbuff[i] = 0;
        np->tx_ring[i].tx_status = 0;
        np->tx_ring[i].desc_length = cpu_to_le32(0x00e08000);
        np->tx_ring[i].next_desc = virt_to_le32desc(&np->tx_ring[i+1]);
        np->tx_buf[i] = kmalloc(PKT_BUF_SZ, GFP_KERNEL);
    }
    np->tx_ring[i-1].next_desc = virt_to_le32desc(&np->tx_ring[0]);

    return;
}


static int start_tx(struct sk_buff *skb, struct net_device *dev)
{
    netdev_private *np = (netdev_private *)dev->priv;
    unsigned entry;

    /* Block a timer-based transmit from overlapping.  This could better be
       done with atomic_swap(1, dev->tbusy), but set_bit() works as well. */

# if(LINUX_VERSION_CODE >= 0x02032a)
    if (netif_queue_stopped(dev)) {
        netif_stop_queue(dev);
        if (jiffies - dev->trans_start > TX_TIMEOUT)
            tx_timeout(dev);
        return 1;
    }
    else
        netif_stop_queue(dev);
#else
    if (test_and_set_bit(0, (void*)&dev->tbusy) != 0) {
        /* This watchdog code is redundant with the media monitor timer. */
        if (jiffies - dev->trans_start > TX_TIMEOUT)
            tx_timeout(dev);
        return 1;
    }
#endif
    /* Explicitly flush packet data cache lines here. */

    /* Caution: the write order is important here, set the descriptor word
       with the "ownership" bit last.  No SMP locking is needed if the
       cur_tx is incremented after the descriptor is consistent.  */

    /* Calculate the next Tx descriptor entry. */
    entry = np->cur_tx % TX_RING_SIZE;

    np->tx_skbuff[entry] = skb;
    
    if ((np->drv_flags & ReqTxAlign)  && (long)skb->data & 3) {
		/* Must use alignment buffer. */
		if (np->tx_buf[entry] == NULL &&
			(np->tx_buf[entry] = kmalloc(PKT_BUF_SZ, GFP_KERNEL)) == NULL)
			return 1;
		memcpy(np->tx_buf[entry], skb->data, skb->len);
		np->tx_ring[entry].addr = virt_to_le32desc(np->tx_buf[entry]);
	} else
		np->tx_ring[entry].addr = virt_to_le32desc(skb->data);
    

    np->tx_ring[entry].desc_length =
        cpu_to_le32(0x00E08000 | (skb->len >= ETH_ZLEN ? skb->len : ETH_ZLEN));
    np->tx_ring[entry].tx_status = cpu_to_le32(DescOwn);

    np->cur_tx++;

    /* Explicitly flush descriptor cache lines here. */

    /* Wake the potentially-idle transmit channel. */
    writew(CmdTxDemand | np->chip_cmd, dev->base_addr + ChipCmd);

    if (np->cur_tx - np->dirty_tx < TX_QUEUE_LEN - 1)
#if (LINUX_VERSION_CODE >= 0x02032a)
        netif_wake_queue(dev);
#else
        clear_bit(0, (void*)&dev->tbusy);       /* Typical path */
#endif
    else
        np->tx_full = 1;
    dev->trans_start = jiffies;

    if (debug > 4) {
        printk(KERN_INFO "%s: Transmit frame #%d queued in slot %d.\n",
               dev->name, np->cur_tx, entry);
    }
    return 0;
}

/* The interrupt handler does all of the Rx thread work and cleans up
   after the Tx thread. */
static void intr_handler(int irq, void *dev_instance, struct pt_regs *rgs)
{
    struct net_device *dev = (struct net_device *)dev_instance;
    netdev_private *np = (void *)dev->priv;
    long ioaddr = dev->base_addr;
    int boguscnt = max_interrupt_work;

    do {
        u32 intr_status = readw(ioaddr + IntrStatus);

        /* Acknowledge all of the current interrupt sources ASAP. */
        writew(intr_status & 0xffff, ioaddr + IntrStatus);

        if (debug > 4)
            printk(KERN_INFO "%s: Interrupt, status %4.4x.\n",
                   dev->name, intr_status);

        if (intr_status == 0)
            break;

        if (intr_status & (IntrRxDone | IntrRxErr | IntrRxDropped |
                           IntrRxWakeUp | IntrRxEmpty | IntrRxNoBuf))
            netdev_rx(dev);

        for (; np->cur_tx - np->dirty_tx > 0; np->dirty_tx++) {
            int entry = np->dirty_tx % TX_RING_SIZE;
            int txstatus = le32_to_cpu(np->tx_ring[entry].tx_status);
            if (txstatus & DescOwn)
                break;
            if (debug > 6)
                printk(KERN_INFO " Tx scavenge %d status %4.4x.\n",
                       entry, txstatus);
            if (txstatus & 0x8000) {
                if (debug > 1)
                    printk(KERN_INFO "%s: Transmit error, Tx status %4.4x.\n",
                           dev->name, txstatus);
                np->stats.tx_errors++;
                
                if (txstatus & 0x0800) { 
                    /* uderrun happen */
                    np->tx_ring[entry].tx_status = cpu_to_le32(DescOwn);
                    writel(virt_to_bus(&np->tx_ring[entry]), ioaddr + TxRingPtr);      
                    /* Turn on Tx On*/
                    writew(CmdTxOn | np->chip_cmd, dev->base_addr + ChipCmd);        
                    /* Stats counted in Tx-done handler, just restart Tx. */
                    writew(CmdTxDemand | np->chip_cmd, dev->base_addr + ChipCmd);        
                    if (debug > 1) 
                       printk(KERN_ERR "Underrun happen");
                    break;
                }    
                if (txstatus & 0x0400) np->stats.tx_carrier_errors++;
                if (txstatus & 0x0200) np->stats.tx_window_errors++;
                if (txstatus & 0x0100) { 
                    np->stats.tx_aborted_errors++;
                    np->tx_ring[entry].tx_status = cpu_to_le32(DescOwn);
                    writel(virt_to_bus(&np->tx_ring[entry]), ioaddr + TxRingPtr);      
                    /* Turn on Tx On*/
                    writew(CmdTxOn | np->chip_cmd, dev->base_addr + ChipCmd);        
                    /* Stats counted in Tx-done handler, just restart Tx. */
                    writew(CmdTxDemand | np->chip_cmd, dev->base_addr + ChipCmd);        
                    //if (debug > 1) 
                    //   printk(KERN_ERR "Tx Abort");
                    break; 
                }    
                if (txstatus & 0x0080) np->stats.tx_heartbeat_errors++;
                if (txstatus & 0x0002) np->stats.tx_fifo_errors++;
#ifdef ETHER_STATS
                if (txstatus & 0x0100) np->stats.collisions16++;
#endif
                /* Transmitter restarted in 'abnormal' handler. */
            } else {
#ifdef ETHER_STATS
                if (txstatus & 0x0001) np->stats.tx_deferred++;
#endif
                np->stats.collisions += (txstatus >> 3) & 15;
#if defined(NETSTATS_VER2)
                np->stats.tx_bytes += np->tx_skbuff[entry]->len;
#endif
                np->stats.tx_packets++;
            }
            /* Free the original skb. */
            dev_free_skb(np->tx_skbuff[entry]);
            np->tx_skbuff[entry] = 0;
        }
#if (LINUX_VERSION_CODE >= 0x02032a)        
        if (np->tx_full && netif_queue_stopped(dev)
            && np->cur_tx - np->dirty_tx < TX_QUEUE_LEN - 4) {
#else
        if (np->tx_full && dev->tbusy
            && np->cur_tx - np->dirty_tx < TX_QUEUE_LEN - 4) {
#endif
            /* The ring is no longer full, clear tbusy. */
            np->tx_full = 0;
#if (LINUX_VERSION_CODE >= 0x02032a)
            netif_wake_queue(dev);
#else
            clear_bit(0, (void*)&dev->tbusy);
            mark_bh(NET_BH);
#endif
        }

        /* Abnormal error summary/uncommon events handlers. */
        if (intr_status & (IntrPCIErr | IntrLinkChange | IntrMIIChange |
                           IntrStatsMax | IntrTxAbort | IntrTxUnderrun))
            netdev_error(dev, intr_status);

        if (--boguscnt < 0) {
            if (debug > 1) 
                printk(KERN_INFO "%s: Too much work at interrupt, "
                       "status=0x%4.4x.\n",
                   dev->name, intr_status);
            break;
        }
    } while (1);

    if (debug > 3)
        printk(KERN_INFO "%s: exiting interrupt, status=%#4.4x.\n",
               dev->name, readw(ioaddr + IntrStatus));

    return;
}

/* This routine is logically part of the interrupt handler, but isolated
   for clarity and better register allocation. */
static int netdev_rx(struct net_device *dev)
{
    netdev_private *np = (netdev_private *)dev->priv;
    int entry = np->cur_rx % RX_RING_SIZE;
    int boguscnt = np->dirty_rx + RX_RING_SIZE - np->cur_rx;

    if (debug > 4) {
        printk(KERN_INFO " In netdev_rx(), entry %d status %8.8x.\n",
               entry, np->rx_head_desc->rx_status);
    }

    /* If EOP is set on the next entry, it's a new packet. Send it up. */
    while ( ! (np->rx_head_desc->rx_status & cpu_to_le32(DescOwn))) {
        rx_desc *desc = np->rx_head_desc;
        u32 desc_status = le32_to_cpu(desc->rx_status);
        int data_size = desc_status >> 16;

        if (debug > 4)
            printk(KERN_INFO "  netdev_rx() status is %4.4x.\n",
                   desc_status);
        if (--boguscnt < 0)
            break;
        if ( (desc_status & (RxWholePkt | RxErr)) !=  RxWholePkt) {
            if ((desc_status & RxWholePkt) !=  RxWholePkt) {
                printk(KERN_INFO "%s: Oversized Ethernet frame spanned "
                       "multiple buffers, entry %#x length %d status %4.4x!\n",
                       dev->name, np->cur_rx, data_size, desc_status);
                printk(KERN_INFO "%s: Oversized Ethernet frame %p vs %p.\n",
                       dev->name, np->rx_head_desc,
                       &np->rx_ring[np->cur_rx % RX_RING_SIZE]);
                np->stats.rx_length_errors++;
            } else if (desc_status & RxErr) {
                /* There was a error. */
                if (debug > 2)
                    printk(KERN_INFO "  netdev_rx() Rx error was %8.8x.\n",
                           desc_status);
                np->stats.rx_errors++;
                if (desc_status & 0x0030) np->stats.rx_length_errors++;
                if (desc_status & 0x0048) np->stats.rx_fifo_errors++;
                if (desc_status & 0x0004) np->stats.rx_frame_errors++;
                if (desc_status & 0x0002) np->stats.rx_crc_errors++;
            }
        } else {
            struct sk_buff *skb;
            /* Length should omit the CRC */
            int pkt_len = data_size - 4;
            unsigned int wLen = 0, wSAP = 0, wActualLen = 0;

            /* For conforming IEEE 802.3 spec
             * If the incoming packet is IEE 802.3 frmae/IEEE 802.3 SNAP frame, get
             * RX_Length in RDES0 from the incoming packet, subtract Ethernet header
             * length and CRC length from it. Then, compare the result with L/T field
             * of the packet. If they're not equal, descard this packet.
             */

            wLen = (*(np->rx_skbuff[entry]->tail + 12) << 8) + (*(np->rx_skbuff[entry]->tail + 13));
            if ( wLen >= 46 && wLen <= 1500) {   // IEEE 802.3/IEEE 802.3 SNAP frame
    	        wSAP = (*(np->rx_skbuff[entry]->tail + 14) << 8) + (*(np->rx_skbuff[entry]->tail + 15));
                if (wSAP != 0xFFFF)               // exclude Novell's Ethernet 802.3 frame
                    wActualLen = data_size - U_HEADER_LEN - U_CRC_LEN;         // real packet length
            }
            if (!(wLen >= 46 && wLen <= 1500 && wSAP != 0xFFFF && wLen != wActualLen)) {
                /* Check if the packet is long enough to accept without copying
                   to a minimally-sized skbuff. */
                if (pkt_len < rx_copybreak
                    && (skb = dev_alloc_skb(pkt_len + 2)) != NULL) {
                    skb->dev = dev;
                    skb_reserve(skb, 2);    /* 16 byte align the IP header */
#if HAS_IP_COPYSUM          /* Call copy + cksum if available. */
                    eth_copy_and_sum(skb, np->rx_skbuff[entry]->tail, pkt_len, 0);
                    skb_put(skb, pkt_len);
#else
                    memcpy(skb_put(skb, pkt_len), np->rx_skbuff[entry]->tail,
                           pkt_len);
#endif
                } else {
                    skb_put(skb = np->rx_skbuff[entry], pkt_len);
                    np->rx_skbuff[entry] = NULL;
                }
                skb->protocol = eth_type_trans(skb, dev);

                netif_rx(skb);
                dev->last_rx = jiffies;
#if defined(NETSTATS_VER2)
                np->stats.rx_bytes += skb->len;
#endif
                np->stats.rx_packets++;
            }
        }
        entry = (++np->cur_rx) % RX_RING_SIZE;
        np->rx_head_desc = &np->rx_ring[entry];
    }

    /* Refill the Rx ring buffers. */
    for (; np->cur_rx - np->dirty_rx > 0; np->dirty_rx++) {
        struct sk_buff *skb;
        entry = np->dirty_rx % RX_RING_SIZE;
        if (np->rx_skbuff[entry] == NULL) {
            skb = dev_alloc_skb(np->rx_buf_sz);
            np->rx_skbuff[entry] = skb;
            if (skb == NULL)
                break;          /* Better luck next round. */
            skb->dev = dev;         /* Mark as being used by this device. */
            np->rx_ring[entry].addr = virt_to_le32desc(skb->tail);
        }
        np->rx_ring[entry].rx_status = cpu_to_le32(DescOwn);
    }

    /* Pre-emptively restart Rx engine. */
    writew(CmdRxDemand | np->chip_cmd, dev->base_addr + ChipCmd);
    return 0;
}

static void netdev_error(struct net_device *dev, int intr_status)
{
    netdev_private *np = (netdev_private *)dev->priv;
    long ioaddr = dev->base_addr;
    
    if (intr_status & (IntrMIIChange | IntrLinkChange)) {
                
        if (readb(ioaddr + MIIStatus) & 0x02) {
			/* Link failed, restart autonegotiation. */
			if (np->drv_flags & HasDavicomPhy)
				mdio_write(dev, np->phys[0], 0, 0x3300);
		} else
				check_duplex(dev);
        if (debug > 5)
            printk(KERN_ERR "%s: MII status changed: Autonegotiation "
                   "advertising %4.4x  partner %4.4x.\n", dev->name,
               mdio_read(dev, np->phys[0], 4),
               mdio_read(dev, np->phys[0], 5)
               );            
    }
    if (intr_status & IntrStatsMax) {
        np->stats.rx_crc_errors += readw(ioaddr + RxCRCErrs);
        np->stats.rx_missed_errors  += readw(ioaddr + RxMissed);
        writel(0, ioaddr + RxMissed);
    }
    if (intr_status & IntrTxAbort) {
             
       
    }
    if (intr_status & IntrTxUnderrun) {
        if (np->tx_thresh < 0xE0)
            writeb(np->tx_thresh += 0x20, ioaddr + TxConfig);        
            
        if (debug > 1)
            printk(KERN_INFO "%s: Transmitter underrun, increasing Tx "
                   "threshold setting to %2.2x.\n", dev->name, np->tx_thresh);
    }
    if ((intr_status & ~(IntrLinkChange | IntrStatsMax |
                         IntrTxAbort|IntrTxAborted))) {
        if(debug > 3)                                                
        printk(KERN_ERR "%s: Something Wicked happened! %4.4x.\n",
               dev->name, intr_status);
        /* Recovery for other fault sources not known. */
        writew(CmdTxDemand | np->chip_cmd, dev->base_addr + ChipCmd);
    }
       
}

static struct net_device_stats *get_stats(struct net_device *dev)
{
    netdev_private *np = (netdev_private *)dev->priv;
    long ioaddr = dev->base_addr;

    /* Nominally we should lock this segment of code for SMP, although
       the vulnerability window is very small and statistics are
       non-critical. */
    np->stats.rx_crc_errors += readw(ioaddr + RxCRCErrs);
    np->stats.rx_missed_errors  += readw(ioaddr + RxMissed);
    writel(0, ioaddr + RxMissed);

    return &np->stats;
}

/* The big-endian AUTODIN II ethernet CRC calculation.
   N.B. Do not use for bulk data, use a table-based routine instead.
   This is common code and should be moved to net/core/crc.c */
static unsigned const ethernet_polynomial = 0x04c11db7U;
static inline u32 ether_crc(int length, unsigned char *data)
{
    int crc = -1;

    while(--length >= 0) {
        unsigned char current_octet = *data++;
        int bit;
        for (bit = 0; bit < 8; bit++, current_octet >>= 1) {
            crc = (crc << 1) ^
                ((crc < 0) ^ (current_octet & 1) ? ethernet_polynomial : 0);
        }
    }
    return crc;
}

static void set_rx_mode(struct net_device *dev)
{
    netdev_private *np = (netdev_private *)dev->priv;
    long ioaddr = dev->base_addr;
    u32 mc_filter[2];           /* Multicast hash filter */
    u8 rx_mode;                 /* Note: 0x02=accept runt, 0x01=accept errs */

    if (dev->flags & IFF_PROMISC) {         /* Set promiscuous. */
        /* Unconditionally log net taps. */
        printk(KERN_NOTICE "%s: Promiscuous mode enabled.\n", dev->name);
        rx_mode = 0x1C;
    } else if ((dev->mc_count > multicast_filter_limit)
               ||  (dev->flags & IFF_ALLMULTI)) {
        /* Too many to match, or accept all multicasts. */
        writel(0xffffffff, ioaddr + MulticastFilter0);
        writel(0xffffffff, ioaddr + MulticastFilter1);
        rx_mode = 0x0C;
    } else {
        struct dev_mc_list *mclist;
        int i;
        memset(mc_filter, 0, sizeof(mc_filter));
        for (i = 0, mclist = dev->mc_list; mclist && i < dev->mc_count;
             i++, mclist = mclist->next) {
            set_bit(ether_crc(ETH_ALEN, mclist->dmi_addr) >> 26,
                    mc_filter);
        }
        writel(mc_filter[0], ioaddr + MulticastFilter0);
        writel(mc_filter[1], ioaddr + MulticastFilter1);
        rx_mode = 0x0C;
    }
    writeb(np->rx_thresh | rx_mode, ioaddr + RxConfig);
}

static int mii_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
	u16 *data = (u16 *)&rq->ifr_data;

	switch(cmd) {
	case SIOCGMIIPHY:		/* Get the address of the PHY in use. */
		data[0] = ((netdev_private *)dev->priv)->phys[0] & 0x1f;
		/* Fall Through */
	case SIOCGMIIREG:		/* Read the specified MII register. */
		data[3] = mdio_read(dev, data[0] & 0x1f, data[1] & 0x1f);
		return 0;
	case SIOCSMIIREG:		/* Write the specified MII register */
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		mdio_write(dev, data[0] & 0x1f, data[1] & 0x1f, data[2]);
		return 0;
	default:
		return -EOPNOTSUPP;
	}
}

static int netdev_close(struct net_device *dev)
{
    long ioaddr = dev->base_addr;
    netdev_private *np = (netdev_private *)dev->priv;
    int i;
    int ww;

#if (LINUX_VERSION_CODE >= 0x02032a)
    netif_stop_queue(dev);
#else
    dev->start = 0;
    dev->tbusy = 1;
#endif

    if (debug > 1)
        printk(KERN_INFO "%s: Shutting down ethercard, status was %4.4x.\n",
               dev->name, readw(ioaddr + ChipCmd));

    /* Disable interrupts by clearing the interrupt mask. */
    writew(0x0000, ioaddr + IntrEnable);
    /* if the chip is 3065, we must patch shutdown bug*/
    
    if (np->drv_flags & HasWOL) { 
        
        //Nic Loop Back On
        writeb(readb(ioaddr + TxConfig) | 0x01, ioaddr + TxConfig);
        
        
       //Tx Off
        writeb(readb(ioaddr + ChipCmd) & 0xEF, ioaddr + ChipCmd);
        
        // W_MAX_TIMEOUT is the timeout period
        for (ww = 0; ww < W_MAX_TIMEOUT; ww++) {
            if((readb(ioaddr + ChipCmd) & 0x10) == 0)
                break; 
        } 
        
        //Rx Off
        writeb(readb(ioaddr + ChipCmd) & 0xF7, ioaddr + ChipCmd);
        
        // W_MAX_TIMEOUT is the timeout period
        for (ww = 0; ww < W_MAX_TIMEOUT; ww++) {
            if((readb(ioaddr + ChipCmd) & 0x08) == 0)
                break; 
        } 
        
        if (ww == W_MAX_TIMEOUT) {
    
            // Turn on fifo test
            writew(readw(ioaddr + GFIFOTest) | 0x0001, ioaddr + GFIFOTest);
            // Turn on fifo reject
            writew(readw(ioaddr + GFIFOTest) | 0x0800, ioaddr + GFIFOTest);
            // Turn off fifo test
            writew(readw(ioaddr + GFIFOTest) & 0xFFFE, ioaddr + GFIFOTest);
        }
        
        //Nic Loop Back Off
        writeb(readb(ioaddr + TxConfig) & 0xFE, ioaddr + TxConfig);
    }
    
    /* Stop the chip's Tx and Rx processes. */
    writew(CmdStop, ioaddr + ChipCmd);
    del_timer(&np->timer);

    free_irq(dev->irq, dev);

    /* Free all the skbuffs in the Rx queue. */
    for (i = 0; i < RX_RING_SIZE; i++) {
        np->rx_ring[i].rx_status = 0;
        np->rx_ring[i].addr = 0xBADF00D0; /* An invalid address. */
        if (np->rx_skbuff[i]) {
#if LINUX_VERSION_CODE < 0x20100
            np->rx_skbuff[i]->free = 1;
#endif
            dev_free_skb(np->rx_skbuff[i]);
        }
        np->rx_skbuff[i] = 0;
    }
    for (i = 0; i < TX_RING_SIZE; i++) {
        if (np->tx_skbuff[i])
            dev_free_skb(np->tx_skbuff[i]);
        np->tx_skbuff[i] = 0;
        if (np->tx_buf[i]) {
            kfree(np->tx_buf[i]);
            np->tx_buf[i] = 0;
        }
    }
    MOD_DEC_USE_COUNT;
    return 0;
}
#ifndef MODULE   
int via_rhine_probe(struct net_device *dev)
{
    printk(KERN_INFO "%s", version1);
	return pci_drv_register(&via_rhine_drv_id, dev);
}
#endif   

#ifdef MODULE
int init_module(void)
{
    if (debug)                  /* Emit version even if no cards detected. */
        printk(KERN_INFO "%s", version);
    return pci_drv_register(&via_rhine_drv_id, NULL);
}

void cleanup_module(void)
{
    struct net_device *next_dev;

    pci_drv_unregister(&via_rhine_drv_id);

    /* No need to check MOD_IN_USE, as sys_delete_module() checks. */
    while (root_net_dev) {
        netdev_private *np = (void *)(root_net_dev->priv);
        unregister_netdev(root_net_dev);
#ifdef USE_IO_OPS
        release_region(root_net_dev->base_addr, pci_tbl[np->chip_id].io_size);
#else
        iounmap((char *)(root_net_dev->base_addr));
#endif
        next_dev = np->next_module;
        if (np->priv_addr)
            kfree(np->priv_addr);
        kfree(root_net_dev);
        root_net_dev = next_dev;
    }
}

#endif  /* MODULE */
