#ifndef __KERNEL__
#define __KERNEL__
#endif
/* Include files, designed to support most kernel versions 2.0.0 and later. */
#include <linux/config.h>

#if defined(CONFIG_SMP) && ! defined(__SMP__)
#define __SMP__
#endif

#if defined(MODULE) && defined(CONFIG_MODVERSIONS) && ! defined(MODVERSIONS)
#define MODVERSIONS
/* Older kernels do not include this automatically. */
#endif

#include <linux/version.h>
#if (LINUX_VERSION_CODE < 0x20341) &&  defined(MODVERSIONS)
#include <linux/modversions.h>
#endif

#include <linux/module.h>


#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/pci.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/malloc.h>
#include <linux/interrupt.h>

#include <linux/string.h>
#include <linux/if.h>
#include <asm/io.h>
#include <asm/processor.h>      /* Processor type for cache alignment. */
#include <asm/bitops.h>

#if (LINUX_VERSION_CODE < 0x20155)
#include <linux/bios32.h>
/* A minimal version of the 2.2.* PCI support that handles configuration
   space access.
   Drivers that actually use pci_dev fields must do explicit compatibility.
   Note that the struct pci_dev * "pointer" is actually a byte mapped integer!
*/
#if (LINUX_VERSION_CODE < 0x20020)
typedef struct _pci_dev {
    int not_used;
} pci_dev;
#endif

#define pci_find_slot(bus, devfn) (struct pci_dev*)((bus<<8) | devfn | 0xf0000)
#define bus_number(pci_dev) ((((int)(pci_dev)) >> 8) & 0xff)
#define devfn_number(pci_dev) (((int)(pci_dev)) & 0xff)

#ifndef CONFIG_PCI
extern inline int pci_present(void)
{ 
    return 0;
}
#else
#define pci_present pcibios_present
#endif

#define pci_read_config_byte(pdev, where, valp)\
    pcibios_read_config_byte(bus_number(pdev), devfn_number(pdev), where, valp)
#define pci_read_config_word(pdev, where, valp)\
    pcibios_read_config_word(bus_number(pdev), devfn_number(pdev), where, valp)
#define pci_read_config_dword(pdev, where, valp)\
    pcibios_read_config_dword(bus_number(pdev), devfn_number(pdev), where, valp)
#define pci_write_config_byte(pdev, where, val)\
    pcibios_write_config_byte(bus_number(pdev), devfn_number(pdev), where, val)
#define pci_write_config_word(pdev, where, val)\
    pcibios_write_config_word(bus_number(pdev), devfn_number(pdev), where, val)
#define pci_write_config_dword(pdev, where, val)\
    pcibios_write_config_dword(bus_number(pdev), devfn_number(pdev), where, val)
#endif

/* Keep the ring sizes a power of two for compile efficiency.
   The compiler will convert <unsigned>'%'<2^N> into a bit mask.
   Making the Tx ring too large decreases the effectiveness of channel
   bonding and packet priority.
   There are no ill effects from too-large receive rings. */
#define TX_RING_SIZE	16
#define TX_QUEUE_LEN	10      /* Limit ring entries actually used.  */
#define RX_RING_SIZE	16
#define U_HEADER_LEN	14
#define U_CRC_LEN	4
#define PCI_REG_MODE3	0x53
#define MODE3_MIION	0x04    /* in PCI_REG_MOD3 OF PCI space */
#define PRIV_ALIGN	15  /* Required alignment mask */

/* Operational parameters that usually are not changed. */
/* Time in jiffies before concluding the transmitter is hung. */
#define TX_TIMEOUT	(2*HZ)
#define PKT_BUF_SZ	1900                 /* Size of each temporary Rx buffer.*/


/* These are used by the netdrivers to report values from the
   MII (Media Indpendent Interface) management registers.
*/
#ifndef SIOCGMIIPHY
//warning You are using an old kern_compat.h version.
#define SIOCGMIIPHY	(SIOCDEVPRIVATE)		/* Get the PHY in use. */
#define SIOCGMIIREG	(SIOCDEVPRIVATE + 1) 		/* Read a PHY register. */
#define SIOCSMIIREG	(SIOCDEVPRIVATE + 2) 		/* Write a PHY register. */
#define SIOCGPARAMS	(SIOCDEVPRIVATE + 3) 		/* Read operational parameters. */
#define SIOCSPARAMS	(SIOCDEVPRIVATE + 4) 		/* Set operational parameters. */
#endif

/* Condensed bus+endian portability operations. */
#define virt_to_le32desc(addr)	cpu_to_le32(virt_to_bus(addr))
#define le32desc_to_virt(addr)	bus_to_virt(le32_to_cpu(addr))

#define USE_IO_OPS
#undef readb
#undef readw
#undef readl
#undef writeb
#undef writew
#undef writel
#define readb	inb
#define readw	inw
#define readl	inl
#define writeb	outb
#define writew	outw
#define writel	outl

// max time out delay time
#define W_MAX_TIMEOUT	0x0FFFU
	
#define RHINE_IOTYPE	(PCI_USES_IO  | PCI_USES_MASTER | PCI_ADDR0)
#define RHINEII_IOSIZE	256

#if (LINUX_VERSION_CODE < 0x20100)
#define PCI_CAPABILITY_LIST	0x34	/* Offset of first capability list entry */
#define PCI_STATUS_CAP_LIST	0x10	/* Support Capability List */
#define PCI_CAP_ID_PM	0x01	/* Power Management */
#endif

/* SMP and better multiarchitecture support were added.
   Using an older kernel means we assume a little-endian uniprocessor.
*/
#if (LINUX_VERSION_CODE < 0x20123)
#define hard_smp_processor_id()	smp_processor_id()
#define test_and_set_bit(val, addr)	set_bit(val, addr)
#define le16_to_cpu(val)	(val)
#define cpu_to_le16(val)	(val)
#define le16_to_cpus(val)
#define le32_to_cpu(val)	(val)
#define cpu_to_le32(val)	(val)
typedef long	spinlock_t;
#define SPIN_LOCK_UNLOCKED	0
#define spin_lock(lock)
#define spin_unlock(lock)
#define spin_lock_irqsave(lock, flags)	save_flags(flags); cli();
#define spin_unlock_irqrestore(lock, flags)	restore_flags(flags);
#endif

#if (LINUX_VERSION_CODE <= 0x20139)
#define	net_device_stats	enet_statistics
#else
#define	NETSTATS_VER2
#endif

/* There was no support for PCI address space mapping in 2.0, but the
   Alpha needed it.  See the 2.2 documentation. */
#if (LINUX_VERSION_CODE < 0x20100) && (!defined(__alpha__))
#define ioremap(a,b)	(((unsigned long)(a) >= 0x100000) ? vremap(a,b) : (void*)(a))
#define iounmap(v)	do { if ((unsigned long)(v) >= 0x100000) vfree(v);} while (0)
#endif

/* The arg count changed, but function name did not.
   We cover that bad choice by defining a new name.
*/
#if (LINUX_VERSION_CODE > 0x020343)
#define dev_free_skb(skb)	dev_kfree_skb_irq(skb)
#elif (LINUX_VERSION_CODE < 0x020159)
#define dev_free_skb(skb)	dev_kfree_skb(skb, FREE_WRITE);
#else
#define dev_free_skb(skb)	dev_kfree_skb(skb);
#endif

/* The old 'struct device' used a too-generic name. */
#if (LINUX_VERSION_CODE < 0x2030d)
#define net_device	device
#endif

/* The 2.2 kernels added the start of capability-based security for operations
   that formerally could only be done by root.
*/
#if (!defined(CAP_NET_ADMIN))
#define capable(CAP_XXX)	(suser())
#endif

/* Support for adding info about the purpose of and parameters for kernel
   modules was added in 2.1. */
#if (LINUX_VERSION_CODE < 0x20115)
#define MODULE_AUTHOR(name)	extern int nonesuch
#define MODULE_DESCRIPTION(string)	extern int nonesuch
#define MODULE_PARM(varname, typestring)	extern int nonesuch
#endif

/*
  These are the structures in the table that drives the PCI probe routines.
  Note the matching code uses a bitmask: more specific table entries should
  be placed before "catch-all" entries.

  The table must be zero terminated.
*/
typedef enum _chip_capability_flags {
    CanHaveMII = 1,
    HasESIPhy = 2,
    HasDavicomPhy = 4,
    ReqTxAlign = 0x10,
    HasWOL = 0x20,
} chip_capability_flags;

typedef enum _pci_id_flags_bits {
    /* Set PCI command register bits before calling probe1(). */
    PCI_USES_IO = 1,
    PCI_USES_MEM = 2,
    PCI_USES_MASTER = 4,
    /* Read and map the single following PCI BAR. */
    PCI_ADDR0 = 0 << 4,
    PCI_ADDR1 = 1 << 4,
    PCI_ADDR2 = 2 << 4,
    PCI_ADDR3 = 3 << 4,
    PCI_ADDR_64BITS = 0x100,
    PCI_NO_ACPI_WAKE = 0x200,
    PCI_NO_MIN_LATENCY = 0x400,
} pci_id_flags_bits;

typedef enum _drv_id_flags {
    PCI_HOTSWAP = 1, /* Leave module loaded for Cardbus-like chips. */
} drv_id_flags;

typedef enum _drv_pwr_action {
    DRV_NOOP,			/* No action. */
    DRV_ATTACH,			/* The driver may expect power ops. */
    DRV_SUSPEND,		/* Machine suspending, next event RESUME or DETACH. */
    DRV_RESUME,			/* Resume from previous SUSPEND  */
    DRV_DETACH,			/* Card will-be/is gone. Valid from SUSPEND! */
    DRV_PWR_WakeOn,		/* Put device in e.g. Wake-On-LAN mode. */
    DRV_PWR_DOWN,		/* Go to lowest power mode. */
    DRV_PWR_UP,			/* Go to normal power mode. */
} drv_pwr_action;

typedef enum _register_offsets {
    StationAddr = 0x00,
    RxConfig = 0x06,
    TxConfig = 0x07,
    ChipCmd = 0x08,
    IntrStatus = 0x0C,
    IntrEnable = 0x0E,
    MulticastFilter0 = 0x10,
    MulticastFilter1 = 0x14,
    RxRingPtr = 0x18,
    TxRingPtr = 0x1C,
    GFIFOTest = 0x54,
    MIIPhyAddr = 0x6C,
    MIIStatus = 0x6D,
    PCIBusConfig = 0x6E, 
    MIICmd = 0x70,
    MIIRegAddr = 0x71,
    MIIData = 0x72,
    MACRegEEcsr = 0x74,
    ConfigA = 0x78,
    ConfigD = 0x7B,
    RxMissed = 0x7C,
    RxCRCErrs = 0x7E,
    StickyHW = 0x83,
    WOLcrClr = 0xA4,
    WOLcgClr = 0xA7,
    PwrcsrClr = 0xAC
} register_offsets;

/* Bits in the interrupt status/mask registers. */
typedef enum _intr_status_bits {
    IntrRxDone = 0x0001,
    IntrRxErr = 0x0004,
    IntrRxEmpty = 0x0020,
    IntrTxDone = 0x0002,
    IntrTxAbort = 0x0008,
    IntrTxUnderrun = 0x0010,
    IntrPCIErr = 0x0040,
    IntrStatsMax = 0x0080,
    IntrRxEarly = 0x0100,
    IntrMIIChange = 0x0200,
    IntrRxOverflow = 0x0400,
    IntrRxDropped = 0x0800,
    IntrRxNoBuf = 0x1000,
    IntrTxAborted = 0x2000,
    IntrLinkChange = 0x4000,
    IntrRxWakeUp = 0x8000,
    IntrNormalSummary = 0x0003,
    IntrAbnormalSummary = 0xC260,
} intr_status_bits;

typedef enum _CFGD_bits {
    CRADOM = 0x08,
    CAP = 0x04,
    MBA = 0X02,
    BAKOPT = 0x01
} CFGD_bits;

typedef enum _TxConfig_bits {
    TCR_LB0 = 0x02,
    TCR_LB1 = 0X04,
    TCR_OFSET = 0x08,
    TCR_RTFT0 = 0x20,
    TCR_RTFT1 = 0x40,
    TCR_RTSF = 0x80
} _TxConfig_bits;

/* Bits in *_desc.status */
typedef enum _rx_status_bits {
    RxOK = 0x8000,
    RxWholePkt = 0x0300,
    RxErr = 0x008F
} rx_status_bits;

typedef enum _desc_status_bits {
    DescOwn = 0x80000000,
    DescEndPacket = 0x4000,
    DescIntr = 0x1000,
} desc_status_bits;

/* Bits in ChipCmd. */
typedef enum _chip_cmd_bits {
    CmdInit = 0x0001,
    CmdStart = 0x0002,
    CmdStop = 0x0004,
    CmdRxOn = 0x0008,
    CmdTxOn = 0x0010,
    CmdTxDemand = 0x0020,
    CmdRxDemand = 0x0040,
    CmdEarlyRx = 0x0100,
    CmdEarlyTx = 0x0200,
    CmdFDuplex = 0x0400,
    CmdNoTxPoll = 0x0800,
    CmdReset = 0x8000,
} chip_cmd_bits;

typedef struct _match_info {
    int	pci, pci_mask, subsystem, subsystem_mask;
    int	revision, revision_mask; 				/* Only 8 bits. */
} match_info;

typedef struct _pci_id_info {
    const char	*name;
    match_info	id;
    pci_id_flags_bits	pci_flags;
    int	io_size;				/* Needed for I/O region check or ioremap(). */
    int	drv_flags;				/* Driver use, intended as capability flags. */
} pci_id_info;

typedef struct _drv_id_info {
    const char	*name;			/* Single-word driver name. */
    int	flags;
    int	pci_class;				/* Typically PCI_CLASS_NETWORK_ETHERNET<<8. */
    pci_id_info	*pci_dev_tbl;
    void *(*probe1)(struct pci_dev *pdev, void *dev_ptr, long ioaddr, int irq, int table_idx, int fnd_cnt, u32 revision);
    /* Optional, called for suspend, resume and detach. */
    int (*pwr_event)(void *dev, int event);
} drv_id_info;


/* The Rx and Tx buffer descriptors. */
typedef struct _rx_desc {
    s32	rx_status;
    u32	desc_length;
    u32	addr;
    u32	next_desc;
} rx_desc;

typedef struct _tx_desc {
    s32	tx_status;
    u32	desc_length;
    u32	addr;
    u32	next_desc;
} tx_desc;


typedef struct _netdev_private {
    /* Descriptor rings first for alignment. */
    rx_desc	rx_ring[RX_RING_SIZE];
    tx_desc	tx_ring[TX_RING_SIZE];
    /* The addresses of receive-in-place skbuffs. */
    struct	sk_buff* rx_skbuff[RX_RING_SIZE];
    /* The saved address of a sent-in-place packet/buffer, for later free(). */
    struct	sk_buff* tx_skbuff[TX_RING_SIZE];
    unsigned char	*tx_buf[TX_RING_SIZE];    /* Tx bounce buffers */
    unsigned char	*tx_bufs;             /* Tx bounce buffer region. */
    struct	net_device *next_module;     /* Link for devices of this type. */
    void	*priv_addr;                    /* Unaligned address for kfree */
    struct	net_device_stats stats;
    struct	timer_list timer;    /* Media monitoring timer. */
    /* Frequently used values: keep some adjacent for cache effect. */
    int	chip_id, drv_flags;
    struct	pci_dev *pci_dev;
    rx_desc	*rx_head_desc;
    unsigned int	cur_rx, dirty_rx;      /* Producer/consumer ring indices */
    unsigned int	cur_tx, dirty_tx;
    unsigned int	rx_buf_sz;             /* Based on MTU+slack. */
    u16	chip_cmd;                       /* Current setting for ChipCmd */
    unsigned int	tx_full:1;             /* The Tx queue is full. */
    /* These values are keep track of the transceiver/media in use. */
    unsigned int	full_duplex:1;         /* Full-duplex operation requested. */
    unsigned int	duplex_lock:1;
    unsigned int	medialock:1;           /* Do not sense media. */
    unsigned int	default_port:4;        /* Last dev->if_port value. */
    u8	tx_thresh, rx_thresh;
    /* MII transceiver section. */
    int	mii_cnt;                        /* MII device addresses. */
    u16	advertising;                    /* NWay media advertisement */
    unsigned char	phys[2];              /* MII device addresses. */
} netdev_private;

int (*register_cb_hook)(drv_id_info *did);
void (*unregister_cb_hook)(drv_id_info *did);
static void *via_probe1(struct pci_dev *pdev, void *init_dev, long ioaddr, int irq, int chip_idx, int find_cnt, u32 revision);
static int pci_drv_register(drv_id_info *drv_id, void *initial_device);
static void pci_drv_unregister(drv_id_info *drv_id);
static int pci_find(struct pci_dev *pdev, int findtype);
static int acpi_wake(struct pci_dev *pdev);
static int  mdio_read(struct net_device *dev, int phy_id, int location);
static void mdio_write(struct net_device *dev, int phy_id, int location, int value);
static int  netdev_open(struct net_device *dev);
static void check_duplex(struct net_device *dev);
static void netdev_timer(unsigned long data);
static void tx_timeout(struct net_device *dev);
static void init_ring(struct net_device *dev);
static int  start_tx(struct sk_buff *skb, struct net_device *dev);
static void intr_handler(int irq, void *dev_instance, struct pt_regs *regs);
static int  netdev_rx(struct net_device *dev);
static void netdev_error(struct net_device *dev, int intr_status);
static void set_rx_mode(struct net_device *dev);
static struct net_device_stats *get_stats(struct net_device *dev);
static int mii_ioctl(struct net_device *dev, struct ifreq *rq, int cmd);
static int  netdev_close(struct net_device *dev);
